#!/usr/bin/env python

import logging
import os
import urllib

import flask


def get_files(d, fnfilter, dfilter, rel=True):
    try:
        d = os.path.expanduser(d)
        dirs = []
        fns = []
        for fn in sorted(os.listdir(d)):
            try:
                ffn = os.path.join(d, fn)
                if not rel:
                    fn = ffn
                if os.path.isdir(ffn):
                    if dfilter(ffn):
                        dirs.append(fn)
                else:
                    if fnfilter(ffn):
                        fns.append(fn)
            except:
                pass
        return fns, dirs
    except:
        return [],[]


def make_blueprint(app=None, register=True, fnfilter=None, dfilter=None):
    if fnfilter is None:
        fnfilter = lambda fn: True
    if dfilter is None:
        dfilter = lambda d: True
    main_dir = os.path.dirname(os.path.abspath(__file__))
    template_folder = os.path.join(main_dir, 'templates')
    static_folder = os.path.join(main_dir, 'static')
    logging.debug('filetree main_dir: %s' % main_dir)
    logging.debug('filetree template_folder: %s' % template_folder)
    logging.debug('filetree static_folder: %s' % static_folder)
    filetree = flask.Blueprint('filetree', 'filetree', \
            template_folder=template_folder, static_folder=static_folder)

    @filetree.route('/json')
    def dirlist():
        try:
            d = urllib.parse.unquote(flask.request.args.get('dir', './'))
            fns, dirs = get_files(d, fnfilter, dfilter, rel=False)
            error = ""
        except Exception as E:
            fns = []
            dirs = []
            error = "PY: %s" % E
        return flask.jsonify(fns=fns, dirs=dirs, error=error)

    @filetree.route('/sfiles', methods=['POST'])
    def sfiles():
        r = []
        try:
            d = urllib.parse.unquote(flask.request.form.get('dir', './'))
            if d == "":
                d = app.config["SLIDE_DIR"]
            fns, dirs = get_files(d, fnfilter, dfilter, rel=True)
            r = ['<ul class="jqueryFileTree" style="display: none;">']
            for f in dirs:
                ff = os.path.join(d, f)
                r.append('<li class="directory collapsed">' \
                        '<a href="#" rel="%s/">%s</a></li>' % (ff, f))
            for f in fns:
                ff = os.path.join(d, f)
                e = os.path.splitext(f)[1][1:]  # get .ext and remove dot
                r.append('<li class="file ext_%s">' \
                '<a href="#" rel="%s">%s</a></li>' % (e, ff, f))
            r.append('</ul>')
        except Exception as E:
            r.append('Could not load directory: %s' % (str(E)))
        return ''.join(r)

    @filetree.route('/')
    def test():
        return flask.render_template('filetree_test.html')

    # dirty fix for flask static bug
    @filetree.route('/files/<path:path>')
    def files(path):
        return filetree.send_static_file(path)

    if register:
        if app is None:
            app = flask.Flask('filetree')
        app.register_blueprint(filetree, url_prefix='/filetree')
        return filetree, app
    return filetree


def test(**kwargs):
    ft, app = make_blueprint(register=True)
    logging.debug(app.url_map)
    app.run(**kwargs)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    test()



import logging
try:
    from PyQt5.QtCore import *
    from PyQt5.QtWebEngineWidgets import *
    from PyQt5.QtWidgets import QApplication, QFileDialog, QMessageBox, QPlainTextEdit, QDialog, QSplashScreen, QProgressDialog, QMainWindow, QToolBar, QAction, QStyle
    from PyQt5.QtWebChannel import QWebChannel
    from PyQt5 import QtGui 
    from PyQt5.QtGui import QDesktopServices

except ImportError:
    # dependency missing, issue a warning
    logging.error('dependency not found, please install PyQt5 to enable gui')
    import sys
    sys.exit()

from optparse import OptionParser

import pathlib
from pathlib import Path
from urllib.parse import urlparse
from urllib.parse import parse_qs
from shutil import copyfile
import subprocess

import threading, time
import sys
import socket

import urllib.parse
import urllib.request
import os
import json

# determine if application is a script file or frozen exe
if getattr(sys, 'frozen', False):
    template_folder=os.path.join(sys._MEIPASS, 'templates')
    static_folder=os.path.join(sys._MEIPASS, 'static')
    os.chdir(sys._MEIPASS)
else: #if __file__:
    #template_folder="templates_standalone"
    folderPath = os.path.dirname(pathlib.Path(__file__))
    template_folder=os.path.join(folderPath, 'templates')
    static_folder=os.path.join(folderPath, 'static')
    os.chdir(folderPath)
    
from tissuumaps import views

class CustomWebEnginePage(QWebEnginePage):
    """ Custom WebEnginePage to customize how we handle link navigation """

    def acceptNavigationRequest(self, url,  _type, isMainFrame):
        if _type == QWebEnginePage.NavigationTypeLinkClicked:
            QDesktopServices.openUrl(url)
            return False
        return True
    
    #def javaScriptConsoleMessage(self, level, msg, line, sourceID):
    #    print (level, msg, line, sourceID)
    
class textWindow(QDialog):
    def __init__(self, parent, title, message):
        QDialog.__init__(self, parent)

        self.setMinimumSize(QSize(700, 500))    
        self.setWindowTitle(title) 

        # Add text field
        self.b = QPlainTextEdit(self)
        self.b.setMinimumSize (650,450)
        self.b.setReadOnly(True)
        self.b.insertPlainText(message)
        self.b.move(10,10)
        self.b.resize(400,200)

#DEBUG_PORT = '5588'
#DEBUG_URL = 'http://127.0.0.1:%s' % DEBUG_PORT
#os.environ['QTWEBENGINE_REMOTE_DEBUGGING'] = DEBUG_PORT

class MainWindow(QMainWindow):
    def __init__(self, qt_app, app, *args, **kwargs):
        super(MainWindow, self).__init__()
        self.resize(1400,1000)

        self.browser = webEngine(qt_app, app, self, *args)
        
        self.setCentralWidget(self.browser)
  
        #self.status = QStatusBar()
        #self.setStatusBar(self.status)
        
        self.bar = self.menuBar()
        self.setStyleSheet("""
        QMenuBar {
            border-bottom: 1px solid #911821;
        }
    """)
        file = self.bar.addMenu("File")

        _open = QAction(self.style().standardIcon(QStyle.SP_DialogOpenButton), "Open",self)
        _open.setShortcut("Ctrl+O")
        file.addAction(_open)
        _open.triggered.connect(self.browser.openImage)

        _save = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), "Save project",self)
        _save.setShortcut("Ctrl+S")
        file.addAction(_save)
        def trigger():
            self.browser.page().runJavaScript("flask.standalone.saveProject();")
        _save.triggered.connect(trigger)
        file.addSeparator()

        _export = QAction(self.style().standardIcon(QStyle.SP_DialogSaveButton), "Export to static webpage",self)
        file.addAction(_export)
        def trigger():
            self.browser.page().runJavaScript("flask.standalone.exportToStatic();")
        _export.triggered.connect(trigger)
        
        _exit = QAction(self.style().standardIcon(QStyle.SP_DialogCancelButton), "Exit",self)
        _exit.setShortcut("Ctrl+Q")
        file.addAction(_exit)
        _exit.triggered.connect(self.close)

        plugins = self.bar.addMenu("Plugins")
        for pluginName in app.config["PLUGINS"]:
            _plugin = QAction(pluginName,self)
            plugins.addAction(_plugin)
            def trigger():
                print ("Plugin triggered:", pluginName)
                self.browser.page().runJavaScript("pluginUtils.startPlugin(\""+pluginName+"\");");
            _plugin.triggered.connect(trigger)
        
        about = self.bar.addMenu("About")
        _help = QAction("Help",self)
        about.addAction(_help)
        def trigger():
            QDesktopServices.openUrl(QUrl("https://tissuumaps.github.io/"))
        _help.triggered.connect(trigger)
        _version = QAction("Version",self)
        about.addAction(_version)
        def trigger():
            if getattr(sys, 'frozen', False):
                folderPath = sys._MEIPASS
            else:
                folderPath = os.path.dirname(pathlib.Path(__file__))
            print (os.path.join(folderPath,"VERSION"), os.path.isfile(os.path.join(folderPath,"VERSION")))
            with open(os.path.join(folderPath,"VERSION"), "r") as fh:
                version = fh.read()
            
            QMessageBox.about(self, "Information", "TissUUmaps version " + version)
        _version.triggered.connect(trigger)

        self.showMaximized()
class webEngine(QWebEngineView):
    def __init__(self, qt_app, app, mainWin, args):
        super().__init__()
        self.setAcceptDrops(True)
        self.qt_app = qt_app
        self.app = views.app
        self.args = args
        self.setMinimumSize(800,400)
        self.setContextMenuPolicy(Qt.NoContextMenu)
        self.lastdir = str(Path.home())
        self.setPage(CustomWebEnginePage(self))
        self.webchannel = QWebChannel()
        self.page().setWebChannel(self.webchannel)
        self.webchannel.registerObject('backend', self)
        self.location = None
        self.mainWin = mainWin
        
        self.mainWin.setWindowTitle("TissUUmaps")
        self.mainWin.resize(1024, 800)
        self.setZoomFactor(1.0)
        self.page().profile().clearHttpCache()
        self.page().profile().downloadRequested.connect(
            self.on_downloadRequested
        )
        self.settings().setAttribute(QWebEngineSettings.FullScreenSupportEnabled, True)
        def setfullscreen (request):
            if self.mainWin.windowState() & Qt.WindowFullScreen:
                self.mainWin.showMaximized()
                self.mainWin.bar.setVisible(True)
            else:
                self.mainWin.showFullScreen()
                self.mainWin.bar.setVisible(False)
            request.accept()
        self.page().fullScreenRequested.connect(setfullscreen)

        self.mainWin.setWindowIcon(QtGui.QIcon('static/misc/favicon.ico')) 
        #self.showMaximized()
    
    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()
            links = []
            for url in event.mimeData().urls():
                links.append(str(url.toLocalFile()))
            for link in links:
                #link = link.replace("\\","/").replace(self.app.basedir.replace("\\","/"), "")
                #print ("link",link, self.app.basedir)
                if(".tmap") in link:
                    self.openImagePath(link)
                self.page().runJavaScript(f"flask.standalone.addLayer(\"{link}\");")
            #self.emit(SIGNAL("dropped"), links)
        else:
            event.ignore()

    def on_downloadRequested(self, download):
        old_path = download.path()  # download.path()
        suffix = QFileInfo(old_path).suffix()
        print (suffix)
        path, _ = QFileDialog.getSaveFileName(
            self, "Save File", old_path, "*." + suffix
        )
        if path:
            download.setPath(path)
            download.accept()

            def openImageThread():
                for i in range(100):
                    if os.path.isfile(path):
                        break
                    time.sleep(0.1)
                else:
                    return
                os.startfile(os.path.normpath(path))
            
            threading.Thread(target=openImageThread,daemon=True).start()
        
    def run (self):
        sys.exit(self.qt_app.exec_())

    def setLocation (self, location):
        self.location = location
        while True:
            try:
                if (urllib.request.urlopen(self.location).getcode() == 200):
                    break
            except:
                pass
            
            print ("Impossible to load",self.location)
            time.sleep(0.1)
        print ("loading page ", self.location)
        if (len(self.args) > 0):
            if not self.openImagePath(self.args[0]):
                self.load(QUrl(self.location))
        else:
            self.load(QUrl(self.location))
            
    @pyqtSlot(str)
    def getProperties(self, path):
        try:
            path = urllib.parse.unquote(path)[:-4]
            print (path)
            slide = views._get_slide(path)
            propString = "\n".join([n + ": " + v for n,v in slide.properties.items()])
        except:
            propString = ""
        
        messageBox = textWindow(self,os.path.basename(path) + " properties", propString)
        messageBox.show()
        
    def openImage(self):
        folderpath = QFileDialog.getOpenFileName(self, 'Select a File',self.lastdir)[0]
        self.openImagePath(folderpath)

    @pyqtSlot(str)
    def exportToStatic(self, state):
        imgFiles = []
        otherFiles = []
        def addRelativePath(state, relativePath):
            nonlocal imgFiles, otherFiles
            def addRelativePath_aux (state, path, isImg):
                nonlocal imgFiles, otherFiles
                if not path[0] in state.keys():
                    return
                if len(path) == 1:
                    if isinstance(state[path[0]], list):
                        if isImg:
                            imgFiles += [s for s in state[path[0]]]
                            state[path[0]] = ["data/images/" + os.path.basename(s) for s in state[path[0]]]
                        else: 
                            otherFiles += [relativePath + "/" + s for s in state[path[0]]]
                            state[path[0]] = ["data/files/" + os.path.basename(s) for s in state[path[0]]]
                        
                    else:
                        if isImg:
                            imgFiles += [state[path[0]]]
                            state[path[0]] = "data/images/" + os.path.basename(state[path[0]])
                        else: 
                            otherFiles += [state[path[0]]]
                            state[path[0]] = "data/files/" + os.path.basename(state[path[0]])
                    return
                else:
                    if isinstance(state[path[0]], list):
                        for state_ in state[path[0]]:
                            addRelativePath_aux (state_, path[1:], isImg)
                    else:
                        addRelativePath_aux (state[path[0]], path[1:], isImg)
            
            try:
                relativePath = relativePath.replace("\\","/")
                paths = [
                    ["layers","tileSource"],
                    ["markerFiles","path"],
                    ["regionFiles","path"],
                    ["regionFile"]
                ]
                for path in paths:
                    addRelativePath_aux (state, path, path[0] == "layers")
            except:
                import traceback
                print (traceback.format_exc())
                
            return state
        parsed_url = urlparse(self.url().toString())
        previouspath = parse_qs(parsed_url.query)['path'][0]
        previouspath = os.path.abspath(os.path.join(self.app.basedir, previouspath))

        folderpath = QFileDialog.getExistingDirectory(self, 'Select webpage directory',self.lastdir,
                                                     options=QFileDialog.ShowDirsOnly)
        
        if (not folderpath):
            return {}
        try:
            relativePath = os.path.relpath(previouspath, os.path.dirname(folderpath))
            state = addRelativePath(json.loads(state), relativePath)
            
            with open(folderpath + "/project.tmap", "w") as f:
                json.dump(state, f)
            os.makedirs(os.path.join(folderpath,"data/images"), exist_ok=True )
            os.makedirs(os.path.join(folderpath,"data/files"), exist_ok=True )
            for image in imgFiles:
                image = image.replace(".dzi","")
                views.ImageConverter(os.path.join(previouspath,image), os.path.join(folderpath,"data/images",os.path.basename(image))).convertToDZI()
            for file in otherFiles:
                copyfile(os.path.join(previouspath,file), os.path.join(folderpath,"data/files",os.path.basename(file)))
            import zipfile
            if getattr(sys, 'frozen', False):
                folderPath = sys._MEIPASS
            else:
                folderPath = os.path.dirname(pathlib.Path(__file__))
            with zipfile.ZipFile(os.path.join(folderPath,"web.zip"), 'r') as zip_ref:
                zip_ref.extractall(folderpath)
            print ("Convert Done!")
            #QMessageBox.about(self, "Information", "Export done!")
        except:
            import traceback
            print (traceback.format_exc())

    @pyqtSlot(str)
    def saveProject(self, state):
        def addRelativePath(state, relativePath):
            def addRelativePath_aux (state, path):
                if len(path) == 1:
                    if isinstance(state[path[0]], list):
                        state[path[0]] = [relativePath + "/" + s for s in state[path[0]]]
                    else:
                        state[path[0]] = relativePath + "/" + state[path[0]]
                    return
                if not path[0] in state.keys():
                    return
                else:
                    if isinstance(state[path[0]], list):
                        for state_ in state[path[0]]:
                            addRelativePath_aux (state_, path[1:])
                    else:
                        addRelativePath_aux (state[path[0]], path[1:])
            
            try:
                relativePath = relativePath.replace("\\","/")
                paths = [
                    ["layers","tileSource"],
                    ["markerFiles","path"],
                    ["regionFiles","path"],
                    ["regionFile"]
                ]
                for path in paths:
                    addRelativePath_aux (state, path)
            except:
                import traceback
                print (traceback.format_exc())
                
            return state

        parsed_url = urlparse(self.url().toString())
        previouspath = parse_qs(parsed_url.query)['path'][0]
        previouspath = os.path.abspath(os.path.join(self.app.basedir, previouspath))
        print (previouspath)
        
        folderpath = QFileDialog.getSaveFileName(self, 'Save project as',self.lastdir)[0]
        if (not folderpath):
            return {}

        relativePath = os.path.relpath(previouspath, os.path.dirname(folderpath))
        state = addRelativePath(json.loads(state), relativePath)
        with open(folderpath, "w") as f:
            json.dump(state, f)

    def openImagePath (self, folderpath):
        print ("openImagePath",folderpath)
        try:
            oldBaseDir = self.app.basedir
        except AttributeError:
            oldBaseDir = ""
        self.lastdir = os.path.dirname(folderpath)
        if not folderpath:
            return
        print ("openImagePath",oldBaseDir, folderpath)
        parts = Path(folderpath).parts
        if (not hasattr(self.app, 'cache')):
            setup(self.app)
        self.app.basedir = parts[0]
        imgPath = os.path.join(*parts[1:])
        imgPath = imgPath.replace("\\","/")
        #try:
        #    if not ".tmap" in imgPath:
        #        views._get_slide(imgPath)
        #except:
        #    self.app.basedir = oldBaseDir
        #    import traceback
        #    print (traceback.format_exc())
        #    QMessageBox.about(self, "Error", "TissUUmaps did not manage to open this image.")

        #    return False
        print ("Opening:", self.app.basedir, self.location + imgPath, QUrl(self.location + imgPath))

        filename = os.path.basename(imgPath)
        path = os.path.dirname(imgPath)
        self.load(QUrl(self.location + filename + "?path=" + path))
        self.mainWin.setWindowTitle("TissUUmaps - " + os.path.basename(folderpath))
        return True

    @pyqtSlot()
    def exit(self):
        self.close()
        #sys.exit()

    @pyqtSlot(str, str, result="QJsonObject")
    def addLayer(self, path, folderpath):
        if (folderpath == ""):
            folderpath = QFileDialog.getOpenFileName(self, 'Select a File')[0]
        if not folderpath:
            returnDict = {"dzi":None,"name":None}
            return returnDict
        parts = Path(folderpath).parts
        print (self.app.basedir)
        if (self.app.basedir != parts[0]):
            if (not self.app.basedir == "C:\mnt\data\shared"):
                reply = QMessageBox.question(self, "Error", "All layers must be in the same drive. Would you like to open this image only?")
                reply = reply == QMessageBox.Yes
            else:
                reply = True
            if reply:
                self.openImagePath(folderpath)
            returnDict = {"dzi":None,"name":None}
            return returnDict
        imgPath = os.path.join(*parts[1:])
        try:
            views._get_slide(imgPath)
        except:
            import traceback
            print (traceback.format_exc())
            QMessageBox.about(self, "Error", "TissUUmaps did not manage to open this image.")
            returnDict = {"dzi":None,"name":None}
            return returnDict
        path = os.path.abspath(os.path.join(self.app.basedir, path))
        imgPath = os.path.abspath(os.path.join(self.app.basedir, imgPath))
        print (path, os.path.dirname(imgPath))
        relativePath = os.path.relpath(os.path.dirname(imgPath), path) 
        if ".." in relativePath:
            reply = QMessageBox.question(self, "Error", "Impossible to add layers from a parent folder. Would you like to open this image only?")
            if reply == QMessageBox.Yes:
                self.openImagePath(folderpath)
            returnDict = {"dzi":None,"name":None}
            return returnDict
        returnDict = {
            "dzi":relativePath + "/" + os.path.basename(imgPath) + ".dzi",
            "name":os.path.basename(imgPath)
        }
        print ("returnDict", returnDict)
        return returnDict
    
def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0

def main():
    parser = OptionParser(usage='Usage: %prog [options] [slide-directory]')
    parser.add_option('-B', '--ignore-bounds', dest='DEEPZOOM_LIMIT_BOUNDS',
                default=False, action='store_false',
                help='display entire scan area')
    parser.add_option('-c', '--config', metavar='FILE', dest='config',
                help='config file')
    parser.add_option('-d', '--debug', dest='DEBUG', action='store_true',
                help='run in debugging mode (insecure)')
    parser.add_option('-e', '--overlap', metavar='PIXELS',
                dest='DEEPZOOM_OVERLAP', type='int',
                help='overlap of adjacent tiles [1]')
    parser.add_option('-f', '--format', metavar='{jpeg|png}',
                dest='DEEPZOOM_FORMAT',
                help='image format for tiles [jpeg]')
    parser.add_option('-l', '--listen', metavar='ADDRESS', dest='host',
                default='127.0.0.1',
                help='address to listen on [127.0.0.1]')
    parser.add_option('-p', '--port', metavar='PORT', dest='port',
                type='int', default=5000,
                help='port to listen on [5000]')
    parser.add_option('-Q', '--quality', metavar='QUALITY',
                dest='DEEPZOOM_TILE_QUALITY', type='int',
                help='JPEG compression quality [75]')
    parser.add_option('-s', '--size', metavar='PIXELS',
                dest='DEEPZOOM_TILE_SIZE', type='int',
                help='tile size [254]')
    parser.add_option('-D', '--depth', metavar='LEVELS',
                dest='FOLDER_DEPTH', type='int',
                help='folder depth search for opening files [4]')

    (opts, args) = parser.parse_args()
    # Overwrite only those settings specified on the command line
    for k in dir(opts):
        if not k.startswith('_') and getattr(opts, k) is None:
            delattr(opts, k)
    views.app.config.from_object(opts)
    views.app.config["isStandalone"] = True

    qInstallMessageHandler(lambda x,y,z: None)

    fmt = QtGui.QSurfaceFormat()
    fmt.setColorSpace(QtGui.QSurfaceFormat.sRGBColorSpace);
    fmt.setVersion(4, 1)

    fmt.setProfile(QtGui.QSurfaceFormat.CoreProfile)
    fmt.setSamples(4)
    QtGui.QSurfaceFormat.setDefaultFormat(fmt)

    vp = QtGui.QOpenGLVersionProfile(fmt)
    
    qt_app = QApplication([])

    logo = QtGui.QPixmap('static/misc/design/logo.png')
    logo = logo.scaledToWidth(512, Qt.SmoothTransformation)
    splash = QSplashScreen(logo, Qt.WindowStaysOnTopHint)

    desktop = qt_app.desktop()
    scrn = desktop.screenNumber(QtGui.QCursor.pos())
    currentDesktopsCenter = desktop.availableGeometry(scrn).center()
    splash.move(currentDesktopsCenter - splash.rect().center())

    splash.show()
    #splash.showMessage('Loading TissUUmaps...',Qt.AlignBottom | Qt.AlignCenter,Qt.white)

    qt_app.processEvents()

    port = 5000
    print ("Starting port detection")
    while (is_port_in_use(port)):
        port += 1
        if port == 6000:
            exit(0)
    print ("Ending port detection", port)

    def flaskThread():
        views.app.run(host="127.0.0.1", port=port, threaded=True, debug=False)

    threading.Thread(target=flaskThread,daemon=True).start()

    ui = MainWindow(qt_app, views.app, args)
    ui.browser.setLocation ("http://127.0.0.1:" + str(port) + "/")

    QTimer.singleShot(1000, splash.close)
    ui.browser.run()

if __name__ == '__main__':
    main ()


from tissuumaps import views
from IPython.display import HTML, Javascript, clear_output
from IPython.core.display import display

import threading
import logging
import json
import click
import warnings
import os, time
from pathlib import Path
import uuid


def secho(text, file=None, nl=None, err=None, color=None, **styles):
    pass

def echo(text, file=None, nl=None, err=None, color=None, **styles):
    pass

click.echo = echo
click.secho = secho
class TissUUmapsViewer ():
    def __init__(self, server, image, height=700):
        self.server = server
        self.image = image
        self.id = "tissUUmapsViewer_" + str(uuid.uuid1()).replace("-","")[0:10]
        iframe = ('<iframe src="{src}" style="width: {width}; '
                  'height: {height}; border: none" id="{id}" allowfullscreen></iframe>')
        src = "http://localhost:%d/%s" % (self.server.port, self.image)
        self.htmlIFrame = HTML(iframe.format(width="100%", height=str(height)+"px", src=src, id=self.id))
        display(self.htmlIFrame)
        time.sleep(2)
    
    def screenshot(self):
        screenshot_id = self.id + "_" + str(uuid.uuid1()).replace("-","")[0:6]
        display(Javascript("""
            function listenToMessages_"""+screenshot_id+"""(evt){
                if (evt.data.type != "screenshot") return;
                window.removeEventListener("message", listenToMessages_"""+screenshot_id+""");
                try {
                    IPython.notebook.kernel.execute(`from IPython.display import update_display, HTML`)
                    IPython.notebook.kernel.execute(`obj = HTML("<img src='`+evt.data.img+`'/>")`)
                    IPython.notebook.kernel.execute(`update_display(obj, display_id="display_out_"""+screenshot_id+"""")`)
                } catch (e) { // vscode or jupyterLab can not communicate back...
                    if (e instanceof ReferenceError) {
                        document.getElementById("img_out_"""+screenshot_id+"""").src = evt.data.img;
                        let newNode = document.createElement("span");
                        newNode.innerHTML = "Warning: run viewer.screenshot in a classical Jupyter Notebook if you want the screenshot to be saved.";
                        document.getElementById("img_out_"""+screenshot_id+"""").parentElement.insertBefore(newNode, document.getElementById("img_out_"""+screenshot_id+""""));
                    }
                }
            }
            var iframe = document.getElementById('"""+self.id+"""');
            window.addEventListener("message", listenToMessages_"""+screenshot_id+""");
            iframe.contentWindow.postMessage({'module':'','function':'','arguments':'[]'},'*');
        """))
        display(HTML("<img src='' id='img_out_"+screenshot_id+"'/>"), display_id="display_out_"+screenshot_id)

class TissUUmapsServer ():
    def __init__(self, slideDir, port=5000):
        
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)
        log = logging.getLogger('pyvips')
        log.setLevel(logging.ERROR)
        warnings.filterwarnings('ignore')

        self.started = False
        self.port = port
        self.slideDir = slideDir
        views.app.config['SLIDE_DIR'] = slideDir

        def startServer ():
            views.app.run(host="0.0.0.0", port=self.port, threaded=True, debug=False)
        
        if TissUUmapsServer.is_port_in_use(self.port):
            log.warning (f"Port {self.port} already in use. Impossible to start TissUUmaps server.")
            return

        thread = threading.Thread(target = startServer)
        thread.setdaemon = False
        thread.start()
        self.started = True

    @staticmethod
    def is_port_in_use(port):
        import socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0

    def viewer (self,image, height=700):
        viewer = TissUUmapsViewer(self,image,height)
        return viewer

def opentmap (path, port=5100, height=700):
    path = os.path.abspath(path)
    parts = Path(path).parts
    server = TissUUmapsServer(slideDir=parts[0], port=port)
    return server.viewer(parts[-1]+"?path="+os.path.join(*parts[1:-1]), height)

def loaddata (images=[], csvFiles=[], xSelector="x", ySelector="y", keySelector=None, nameSelector=None, 
              colorSelector=None, piechartSelector=None, shapeSelector=None, scaleSelector=None, 
              fixedShape=None, scaleFactor=1,
              colormap=None,
              compositeMode="source-over",
              boundingBox=None,
              port=5100, height=700, tmapFilename="_project"):
    # make str input to arrays:
    if isinstance(images, str):
        images = [images]
    if isinstance(csvFiles, str):
        csvFiles = [csvFiles]

    # make all paths absolute:
    images = [os.path.abspath(f) for f in images]
    csvFiles = [os.path.abspath(f) for f in csvFiles]
    
    # make all paths relative to project file:
    rootPath = os.path.commonpath(
        [os.path.dirname(f) for f in images] + [os.path.dirname(f) for f in csvFiles]
    )
    images = [os.path.relpath(f, rootPath) for f in images]
    csvFiles = [os.path.relpath(f, rootPath) for f in csvFiles]

    # Create json TMAP file:
    jsonTmap = {
        "compositeMode": compositeMode,
        "filename": "",
        "layers": [
            {
                "name": os.path.basename(layer),
                "tileSource": layer + ".dzi"
            } for layer in images
        ],
        "hideTabs": True,
        "markerFiles": []
    }
    if boundingBox:
        jsonTmap["boundingBox"] = {"x":boundingBox[0],"y":boundingBox[1],"width":boundingBox[2],"height":boundingBox[3]}
    if csvFiles:
        expectedHeader = {
                "X": xSelector,
                "Y": ySelector,
                "gb_col": keySelector,
                "gb_name": nameSelector,
                "cb_cmap": colormap,
                "cb_col": colorSelector,
                "scale_col": scaleSelector,
                "scale_factor": scaleFactor,
                "pie_col": piechartSelector,
                "shape_col": shapeSelector,
                "shape_fixed": fixedShape,
                "cb_gr_dict": "",
                "shape_gr_dict": "",
                "opacity": 1
            }
        expectedRadios = {
                "cb_col": colorSelector!=None,
                "cb_gr": colorSelector==None,
                "cb_gr_rand": False,
                "cb_gr_dict": False,
                "cb_gr_key": True,
                "pie_check": piechartSelector!=None,
                "scale_check": scaleSelector!=None,
                "shape_gr": shapeSelector==None and fixedShape==None,
                "shape_gr_rand": shapeSelector==None and fixedShape==None,
                "shape_gr_dict": False,
                "shape_col": shapeSelector!=None,
                "shape_fixed": fixedShape!=None
            }
        if len(csvFiles) == 1:
            csvFiles = csvFiles[0]
        jsonTmap["markerFiles"] = [{
            "comment": "All markers",
            "path": csvFiles,
            "title": "Download markers",
            "autoLoad": True,
            "hideSettings": True,
            "expectedHeader": expectedHeader,
            "expectedRadios": expectedRadios,
            "uid": "markers"
        }]
    tmapFile = os.path.join(rootPath,f"{tmapFilename}.tmap")
    print ("Creating project file", tmapFile)
    with open(tmapFile, "w") as f:
        json.dump(jsonTmap, f)
    return opentmap (os.path.abspath(tmapFile), port, height)

if __name__ == '__main__':
    pass


#!/usr/bin/env python
#
# flaskTissUUmaps - a minimal python server for TissUUmaps using Flask
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 3.0 of the GNU General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import glob
import os
import pathlib
import sys
from flask import Flask
from optparse import OptionParser
import logging

SLIDE_DIR = "/mnt/data/shared/"
SLIDE_CACHE_SIZE = 60
DEEPZOOM_FORMAT = 'png'
DEEPZOOM_TILE_SIZE = 254
DEEPZOOM_OVERLAP = 1
DEEPZOOM_LIMIT_BOUNDS = True
DEEPZOOM_TILE_QUALITY = 90

FOLDER_DEPTH = 4
PLUGINS = []

READ_ONLY = False

# determine if application is a script file or frozen exe
if getattr(sys, 'frozen', False):
    template_folder=os.path.join(sys._MEIPASS, 'templates')
    static_folder=os.path.join(sys._MEIPASS, 'static')
    plugins_folder=os.path.join(sys._MEIPASS, 'plugins')
    os.chdir(sys._MEIPASS)
else:
    folderPath = os.path.dirname(pathlib.Path(__file__))
    template_folder=os.path.join(folderPath, 'templates')
    static_folder=os.path.join(folderPath, 'static')
    plugins_folder=os.path.join(folderPath, 'plugins')

logging.info("template_folder: " + template_folder)
logging.info("static_folder: " + static_folder)

app = Flask(__name__,template_folder=template_folder,static_folder=static_folder)
app.config.from_object(__name__)
app.config.from_envvar('TISSUUMAPS_CONF', silent=True)
app.config["PLUGIN_FOLDER"] = plugins_folder

for module in glob.glob(app.config["PLUGIN_FOLDER"] + "/*.py"):
    if "__init__.py" in module:
        continue
    app.config["PLUGINS"].append(os.path.splitext(os.path.basename(module))[0])
logging.info("Plugin list:",app.config["PLUGINS"])

app.config["isStandalone"] = False

from . import views



# Python default library
from collections import OrderedDict
from functools import wraps
import gzip
import importlib
import io
import base64
import json
import os
import threading
from threading import Lock
import time
import logging
from urllib.parse import urlparse
from urllib.parse import parse_qs

# External libraries
import imghdr
import pyvips
import openslide
from openslide import (
    ImageSlide,
    OpenSlide,
)
from openslide.deepzoom import DeepZoomGenerator
from tissuumaps import app

# Flask dependencies
from flask import (
    abort,
    make_response,
    render_template,
    url_for,
    request,
    Response,
    send_from_directory,
    redirect,
    _request_ctx_stack
)

from tissuumaps.flask_filetree import filetree
def _fnfilter (filename):
    filename = filename.lower()
    if OpenSlide.detect_format(filename):
        return True
    elif imghdr.what(filename):
        return True
    elif ".tmap" in filename:
        return True
    return False

def _dfilter (filename):
    if "private" in filename:
        return False
    if ".tissuumaps" in filename:
        return False
    return True

ft = filetree.make_blueprint(app=app, register=False, dfilter=_dfilter, fnfilter=_fnfilter)
app.register_blueprint(ft, url_prefix='/filetree')

def check_auth(username, password):
    if username == "username" and password == "password":
        return True
    return False


def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response(
        "Could not verify your access level for that URL.\n"
        "You have to login with proper credentials",
        401,
        {"WWW-Authenticate": 'Basic realm="Login Required"'},
    )


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if _request_ctx_stack.top.request.args.get('path'):
            path = os.path.abspath(os.path.join(app.basedir, _request_ctx_stack.top.request.args.get('path'), "fake"))
        elif not "path" in kwargs.keys():
            path = getPathFromReferrer(_request_ctx_stack.top.request, "")
        else:
            path = os.path.abspath(os.path.join(app.basedir, kwargs["path"]))
        activeFolder = os.path.dirname(path)
        while os.path.dirname(activeFolder) != activeFolder and not os.path.isfile(
            activeFolder + "/auth"
        ):
            activeFolder = os.path.dirname(activeFolder)
        if os.path.isfile(activeFolder + "/auth"):
            with open(activeFolder + "/auth", "r") as file:
                data = file.read().replace("\n", "")
                user, password = [u.strip() for u in data.split(";")]
            auth = request.authorization
            if not auth or not (user == auth.username and password == auth.password):
                return authenticate()
            return f(*args, **kwargs)
        else:
            return f(*args, **kwargs)

    return decorated


class PILBytesIO(io.BytesIO):
    def fileno(self):
        """Classic PIL doesn't understand io.UnsupportedOperation."""
        raise AttributeError("Not supported")


class ImageConverter:
    def __init__(self, inputImage, outputImage):
        self.inputImage = inputImage
        self.outputImage = outputImage

    def convert(self):
        logging.debug(
            "Converting:",
            self.inputImage,
            self.outputImage,
            os.path.isfile(self.outputImage),
        )
        if not os.path.isfile(self.outputImage):

            def convertThread():
                try:
                    imgVips = pyvips.Image.new_from_file(self.inputImage)
                    minVal = imgVips.percent(0.5)
                    maxVal = imgVips.percent(99.5)
                    if minVal == maxVal:
                        minVal = 0
                        maxVal = 255
                    if (imgVips.percent(0.01) < 0 or imgVips.percent(99.99) > 255):
                        imgVips = (255.0 * (imgVips - minVal)) / (maxVal - minVal)
                        imgVips = (imgVips < 0).ifthenelse(0, imgVips)
                        imgVips = (imgVips > 255).ifthenelse(255, imgVips)
                        imgVips = imgVips.scaleimage()
                    imgVips.tiffsave(
                        self.outputImage,
                        pyramid=True,
                        tile=True,
                        tile_width=256,
                        tile_height=256,
                        compression='jpeg',
                        properties=True
                    )
                except:
                    logging.error("Impossible to convert image using VIPS:")
                    import traceback

                    logging.error(traceback.format_exc())
                self.convertDone = True

            self.convertDone = False
            threading.Thread(target=convertThread, daemon=True).start()
            while not self.convertDone:
                time.sleep(0.02)
        return self.outputImage

    def convertToDZI(self):
        if not os.path.isfile(self.outputImage):

            def convertThread():
                try:
                    imgVips = pyvips.Image.new_from_file(self.inputImage)
                    minVal = imgVips.percent(0.5)
                    maxVal = imgVips.percent(99.5)
                    if minVal == maxVal:
                        minVal = 0
                        maxVal = 255
                    if (imgVips.percent(0.01) < 0 or imgVips.percent(99.99) > 255):
                        imgVips = (255.0 * (imgVips - minVal)) / (maxVal - minVal)
                        imgVips = (imgVips < 0).ifthenelse(0, imgVips)
                        imgVips = (imgVips > 255).ifthenelse(255, imgVips)
                        imgVips = imgVips.scaleimage()
                    imgVips.dzsave(
                        os.path.basename(self.outputImage),
                        dirname=os.path.dirname(self.outputImage),
                        suffix='.jpg',
                        background=0,
                        depth='onepixel',
                        overlap=0,
                        tile_size=256
                    )
                    
                except:
                    logging.error("Impossible to convert image using VIPS:")
                    import traceback

                    logging.error(traceback.format_exc())
                self.convertDone = True

            self.convertDone = False
            threading.Thread(target=convertThread, daemon=True).start()
            while not self.convertDone:
                time.sleep(0.02)
        return self.outputImage


class _SlideCache(object):
    def __init__(self, cache_size, dz_opts):
        self.cache_size = cache_size
        self.dz_opts = dz_opts
        self._lock = Lock()
        self._cache = OrderedDict()

    def get(self, path, originalPath=None):
        with self._lock:
            if path in self._cache:
                # Move to end of LRU
                slide = self._cache.pop(path)
                self._cache[path] = slide
                return slide

        osr = OpenSlide(path)
        # try:
        #    osr = OpenSlide(path)
        # except:
        #    osr = ImageSlide(path)
        # Fix for 16 bits tiff files
        # if osr._image.getextrema()[1] > 256:
        #     osr._image = osr._image.point(lambda i:i*(1./256)).convert('L')

        slide = DeepZoomGenerator(osr, **self.dz_opts)
        slide.osr = osr

        slide.associated_images = {}
        for name, image in slide.osr.associated_images.items():
            slide.associated_images[name] = image

        try:
            mpp_x = osr.properties[openslide.PROPERTY_NAME_MPP_X]
            mpp_y = osr.properties[openslide.PROPERTY_NAME_MPP_Y]
            slide.properties = osr.properties
            slide.mpp = (float(mpp_x) + float(mpp_y)) / 2
        except (KeyError, ValueError):
            try:
                if osr.properties["tiff.ResolutionUnit"] == "centimetre":
                    numerator = 10000  # microns in CM
                else:
                    numerator = 25400  # Microns in Inch
                mpp_x = numerator / float(osr.properties["tiff.XResolution"])
                mpp_y = numerator / float(osr.properties["tiff.YResolution"])
                slide.properties = osr.properties
                slide.mpp = (float(mpp_x) + float(mpp_y)) / 2
            except:
                slide.mpp = 0
        try:
            slide.properties = slide.properties
        except:
            slide.properties = osr.properties
        slide.tileLock = Lock()
        if originalPath:
            slide.properties = {"Path":originalPath}
        with self._lock:
            if path not in self._cache:
                while len(self._cache) >= self.cache_size:
                    self._cache.popitem(last=False)
                self._cache[path] = slide
        return slide

class _SlideFile(object):
    def __init__(self, relpath):
        self.name = os.path.basename(relpath)
        self.url_path = relpath.replace("\\", "/")


def setup(app):
    app.basedir = os.path.abspath(app.config["SLIDE_DIR"])
    config_map = {
        "DEEPZOOM_TILE_SIZE": "tile_size",
        "DEEPZOOM_OVERLAP": "overlap",
        "DEEPZOOM_LIMIT_BOUNDS": "limit_bounds",
    }
    opts = dict((v, app.config[k]) for k, v in config_map.items())
    app.cache = _SlideCache(app.config["SLIDE_CACHE_SIZE"], opts)


@app.before_first_request
def _setup():
    setup(app)


@app.errorhandler(404)
def page_not_found(e):
    # note that we set the 404 status explicitly
    #return render_template("tissuumaps.html", isStandalone=app.config["isStandalone"], message="Impossible to load this file", readOnly=app.config["READ_ONLY"])
    return redirect("/404"), 404, {"Refresh": "1; url=/404"}


def _get_slide(path, originalPath=None):
    path = os.path.abspath(os.path.join(app.basedir, path))
    if not path.startswith(app.basedir):
        # Directory traversal
        abort(404)
    if not os.path.exists(path):
        abort(404)
    try:
        slide = app.cache.get(path, originalPath)
        slide.filename = os.path.basename(path)
        return slide
    except:
        if ".tissuumaps" in path:
            abort(404)
        try:
            newpath = (
                os.path.dirname(path)
                + "/.tissuumaps/"
                + os.path.splitext(os.path.basename(path))[0]
                + ".tif"
            )
            if not os.path.isdir(os.path.dirname(path) + "/.tissuumaps/"):
                os.makedirs(os.path.dirname(path) + "/.tissuumaps/")
            tifpath = ImageConverter(path, newpath).convert()
            return _get_slide(tifpath, path)
        except:
            import traceback

            logging.error(traceback.format_exc())
            abort(404)

@app.route("/")
@requires_auth
def index():
    return render_template("tissuumaps.html", isStandalone=app.config["isStandalone"], readOnly=app.config["READ_ONLY"])

@app.route("/web/<path:path>")
@requires_auth
def base_static(path):
    completePath = os.path.abspath(os.path.join(app.basedir, path))
    directory = os.path.dirname(completePath) + "/web/"
    filename = os.path.basename(completePath)
    return send_from_directory(directory, filename)


@app.route("/<path:filename>")
@requires_auth
def slide(filename):
    path = request.args.get('path')
    if not path:
        path = "./"
    path = os.path.abspath(os.path.join(app.basedir, path, filename))
    #slide = _get_slide(path)
    slide_url = os.path.basename(path)+".dzi"#url_for("dzi", path=path)
    jsonProject={
        "layers": [
            {
                "name": os.path.basename(path),
                "tileSource": slide_url
            }
        ]
    }
    return render_template(
        "tissuumaps.html",
        plugins=app.config["PLUGINS"],
        jsonProject=jsonProject,
        isStandalone=app.config["isStandalone"],
        readOnly=app.config["READ_ONLY"]
        )

@app.route("/ping")
@requires_auth
def ping():
    return make_response("pong")

def getPathFromReferrer(request, filename):
    try:
        parsed_url = urlparse(request.referrer)
        path = parse_qs(parsed_url.query)['path'][0]
        path = os.path.abspath(os.path.join(app.basedir, path, filename))
    except:
        path = os.path.abspath(os.path.join(app.basedir, filename))
    if not path:
        path = os.path.abspath(os.path.join(app.basedir, filename))
    logging.debug(f"Path from referrer: {path}")
    return path


@app.route("/<path:path>/<string:filename>.tmap", methods=["GET", "POST"])
@requires_auth
def tmapFile_old(path, filename):
    return redirect(url_for("tmapFile", filename=filename) + "?path=" + path)

@app.route("/<string:filename>.tmap", methods=["GET", "POST"])
@requires_auth
def tmapFile(filename):
    path = request.args.get('path')
    if not path:
        path = "./"
    jsonFilename = os.path.abspath(os.path.join(app.basedir, path, filename) + ".tmap")

    if request.method == "POST" and not app.config["READ_ONLY"]:
        state = request.get_json(silent=False)
        with open(jsonFilename, "w") as jsonFile:
            json.dump(state, jsonFile, indent=4, sort_keys=True)
        return state
    else:
        if os.path.isfile(jsonFilename):
            try:
                with open(jsonFilename, "r") as jsonFile:
                    state = json.load(jsonFile)
            except:
                import traceback

                logging.error(traceback.format_exc())
                abort(404)
        else:
            abort(404)
        if "plugins" in state.keys():
            plugins = state["plugins"]
        else:
            plugins = app.config["PLUGINS"]

        return render_template(
            "tissuumaps.html",
            plugins=plugins,
            jsonProject=state,
            isStandalone=app.config["isStandalone"],
            readOnly=app.config["READ_ONLY"]
        )


@app.route("/<path:completePath>.csv")
@requires_auth
def csvFile(completePath):
    completePath = os.path.join(app.basedir, completePath + ".csv")
    directory = os.path.dirname(completePath)
    filename = os.path.basename(completePath)
    print ("completePath", completePath, directory, filename, os.path.isfile(completePath))
    if os.path.isfile(completePath):
        # Temporary fix for gz files without csv
        if os.path.isfile(completePath + ".gz"):
            os.rename(completePath + ".gz", completePath + ".cgz")

        generate_cgz = False
        if not os.path.isfile(completePath + ".cgz"):
            generate_cgz = True
        elif os.path.getmtime(completePath) > os.path.getmtime(completePath + ".cgz"):
            # In this case, the csv file has been recently modified and the cgz file is
            # stale, so it must be regenerated.
            generate_cgz = True
        if generate_cgz:
            with open(completePath, "rb") as f_in, gzip.open(
                completePath + ".cgz", "wb", compresslevel=9
            ) as f_out:
                f_out.writelines(f_in)

        response = make_response(send_from_directory(directory, filename + ".cgz"))
        response.headers["Content-Encoding"] = "gzip"
        response.headers["Vary"] = "Accept-Encoding"
        response.headers["Transfer-Encoding"] = "gzip"
        response.headers["Content-Length"] = os.path.getsize(completePath + ".cgz")
        response.headers["Content-Type"] = "text/csv; charset=UTF-8"
        return response
    else:
        abort(404)


@app.route("/<path:completePath>.json")
@requires_auth
def jsonFile(completePath):
    completePath = os.path.join(app.basedir, completePath + ".json")
    directory = os.path.dirname(completePath)
    filename = os.path.basename(completePath)
    if os.path.isfile(completePath):
        return send_from_directory(directory, filename)
    else:
        abort(404)


@app.route("/<path:path>.dzi")
@requires_auth
def dzi(path):
    # Check if a .dzi file exists, else use OpenSlide:
    if os.path.isfile(path + ".dzi"):
        directory = os.path.dirname(path)
        filename = os.path.basename(path) + ".dzi"
        return send_from_directory(directory, filename)
    slide = _get_slide(path)
    format = app.config["DEEPZOOM_FORMAT"]
    resp = make_response(slide.get_dzi(format))
    resp.mimetype = "application/xml"
    return resp


@app.route("/<path:path>.dzi/info")
@requires_auth
def dzi_asso(path):
    slide = _get_slide(path)
    associated_images = []
    for key, im in slide.associated_images.items():
        output = io.BytesIO ()
        im.save(output, "PNG")
        b64 = base64.b64encode(output.getvalue()).decode()
        associated_images.append({"name":key,"content":b64})
        output.close()
    return render_template(
            "slide_prop.html",
            associated_images=associated_images,
            properties=slide.properties,
        )
    return resp



@app.route("/<path:path>_files/<int:level>/<int:col>_<int:row>.<format>")
def tile(path, level, col, row, format):
    if os.path.isfile( f"{path}_files/{level}/{col}_{row}.{format}"):
        directory = os.path.dirname(f"{path}_files/{level}/{col}_{row}.{format}")
        filename = os.path.basename(f"{path}_files/{level}/{col}_{row}.{format}")
        return send_from_directory(directory, filename)
    slide = _get_slide(path)
    format = format.lower()
    # if format != 'jpeg' and format != 'png':
    #    # Not supported by Deep Zoom
    #    abort(404)
    try:
        with slide.tileLock:
            tile = slide.get_tile(level, (col, row))
    except ValueError:
        # Invalid level or coordinates
        abort(404)
    buf = PILBytesIO()
    tile.save(buf, format, quality=app.config["DEEPZOOM_TILE_QUALITY"])
    resp = make_response(buf.getvalue())
    resp.mimetype = "image/%s" % format
    resp.cache_control.max_age = 1209600
    resp.cache_control.public = True
    return resp

@app.route(
    "/<path:path>.dzi/<path:associated_name>_files/<int:level>/<int:col>_<int:row>.<format>"
)
def tile_asso(path, associated_name, level, col, row, format):
    slide = _get_slide(path).associated_images[associated_name]
    format = format.lower()
    if format != "jpeg" and format != "png":
        # Not supported by Deep Zoom
        abort(404)
    try:
        tile = slide.get_tile(level, (col, row))
    except ValueError:
        # Invalid level or coordinates
        abort(404)
    buf = PILBytesIO()
    tile.save(buf, format, quality=app.config["DEEPZOOM_TILE_QUALITY"])
    resp = make_response(buf.getvalue())
    resp.mimetype = "image/%s" % format
    return resp


def load_plugin(name):
    try:
        mod = importlib.import_module("." + name, package="tissuumaps.plugins")
    except:
        mod = importlib.import_module("." + name, package="plugins")
    return mod


@app.route("/plugins/<path:pluginName>.js")
def runPlugin(pluginName):
    directory = app.config["PLUGIN_FOLDER"]
    
    filename = pluginName + ".js"
    completePath = os.path.abspath(os.path.join(directory, pluginName + ".js"))
    directory = os.path.dirname(completePath)
    filename = os.path.basename(completePath)
    if os.path.isfile(completePath):
        return send_from_directory(directory, filename)
    else:
        logging.error(completePath, "is not an existing file.")
        abort(404)


@app.route("/plugins/<path:pluginName>/<path:method>", methods=["GET", "POST"])
def pluginJS(pluginName, method):
    logging.info("runPlugin", pluginName, method)
    logging.debug(request.method)

    pluginModule = load_plugin(pluginName)
    pluginInstance = pluginModule.Plugin(app)
    pluginMethod = getattr(pluginInstance, method)
    if request.method == "POST":
        content = request.get_json(silent=False)
        return pluginMethod(content)
    else:
        content = request.args
        return pluginMethod(content)


@app.route("/favicon.ico")
def favicon():
    return send_from_directory(
        os.path.join(app.root_path, "static"),
        "misc/favicon.ico",
        mimetype="image/vnd.microsoft.icon",
    )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("tissuumaps/VERSION", "r") as fh:
    version = fh.read()

setuptools.setup(
     name='TissUUmaps',
     version=version,
     author="Leslie Solorzano, Christophe Avenel, Fredrik Nysjö",
     author_email="christophe.avenel@it.uu.se",
     description="TissUUmaps is a lightweight viewer that uses basic web tools to visualize gene expression data or any kind of point data on top of whole slide images.",
     long_description=long_description,
     long_description_content_type="text/markdown",
     license='GNU General Public License v3.0',
     url="https://tissuumaps.research.it.uu.se/",
     packages=["tissuumaps"],
     include_package_data=True,
     zip_safe=False,
     install_requires=[
        'Flask>=2.0.0',
        'openslide-python>=1.1.2',
        'Pillow>=8.2.0',
        'ipython>=7.28.0',
        'pyvips>=2.1.14'
     ],
     extras_require={
        'pyqt5':[
            'PyQt5>=5.15.4',
            'PyQtWebEngine>=5.15.4'
        ],
        'full':[
            'PyQt5>=5.15.4',
            'PyQtWebEngine>=5.15.4'
        ]
     },
     classifiers=[
         "Programming Language :: Python :: 3",
         'Intended Audience :: Science/Research',
         "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
         "Operating System :: OS Independent",
     ],
     entry_points={
        'console_scripts': [
            'tissuumaps_server = tissuumaps.__main__:main',
            'tissuumaps = tissuumaps.gui:main'
        ]
    }
 )


from optparse import OptionParser

from . import views

def main ():
    parser = OptionParser(usage='Usage: %prog [options] [slide-directory]')
    parser.add_option('-B', '--ignore-bounds', dest='DEEPZOOM_LIMIT_BOUNDS',
                default=False, action='store_false',
                help='display entire scan area')
    parser.add_option('-c', '--config', metavar='FILE', dest='config',
                help='config file')
    parser.add_option('-d', '--debug', dest='DEBUG', action='store_true',
                help='run in debugging mode (insecure)', default=False)
    parser.add_option('-e', '--overlap', metavar='PIXELS',
                dest='DEEPZOOM_OVERLAP', type='int',
                help='overlap of adjacent tiles [1]')
    parser.add_option('-f', '--format', metavar='{jpeg|png}',
                dest='DEEPZOOM_FORMAT',
                help='image format for tiles [jpeg]')
    parser.add_option('-l', '--listen', metavar='ADDRESS', dest='host',
                default='127.0.0.1',
                help='address to listen on [127.0.0.1]')
    parser.add_option('-p', '--port', metavar='PORT', dest='port',
                type='int', default=5000,
                help='port to listen on [5000]')
    parser.add_option('-Q', '--quality', metavar='QUALITY',
                dest='DEEPZOOM_TILE_QUALITY', type='int',
                help='JPEG compression quality [75]')
    parser.add_option('-s', '--size', metavar='PIXELS',
                dest='DEEPZOOM_TILE_SIZE', type='int',
                help='tile size [254]')
    parser.add_option('-D', '--depth', metavar='LEVELS',
                dest='FOLDER_DEPTH', type='int',
                help='folder depth search for opening files [4]')
    parser.add_option('-r', '--readonly', 
                dest='READ_ONLY', action='store_true',
                help='Remove options to save tmap files')

    (opts, args) = parser.parse_args()
    # Load config file if specified
    if opts.config is not None:
        views.app.config.from_pyfile(opts.config)
        
    # Overwrite only those settings specified on the command line
    for k in dir(opts):
        if not k.startswith('_') and getattr(opts, k) is None:
            delattr(opts, k)
    views.app.config.from_object(opts)

    try:
        views.app.config['SLIDE_DIR'] = args[0]
    except IndexError:
        pass

    views.app.run(host=opts.host, port=opts.port, threaded=True, debug=opts.DEBUG)

if __name__ == '__main__':
    main ()
